;; define the utf8 coding system
(make-coding-system 'utf-8 'utf-8
		    "UTF-8 Unicode coding system"
		    '(mnemonic "U8"));
			       
(set-coding-category-system 'utf-8 'utf-8)

;; define the utf16 coding system
(make-coding-system 'utf-16 'utf-16
		    "UTF-16 Unicode coding system (big endian)"
		    '(mnemonic "U16" little-endian nil));
			       
(set-coding-category-system 'utf-16 'utf-16)


(make-coding-system 'utf-16-le 'utf-16
		    "UTF-16 Unicode coding system (little endian)"
		    '(mnemonic "U16-LE" little-endian t));

(define-coding-system-alias 'utf-16-be 'utf-16)
			       
;; a subtle distinction: endian utf16 with boms forbidden, per spec
(make-coding-system 'utf-16le 'utf-16
		    "UTF-16 little-endian Unicode coding system"
		    '(mnemonic "U16LE" bom nil little-endian t))
(make-coding-system 'utf-16be 'utf-16
		    "UTF-16 big-endian Unicode coding system"
		    '(mnemonic "U16BE" bom nil little-endian nil))

;; big and little endian with forced bom
(make-coding-system 'utf-16-le-bom 'utf-16
		    "UTF-16 coding system (little-endian with BOM)"
		    '(mnemonic "U16-LE-B" bom t little-endian t))
(make-coding-system 'utf-16-be-bom 'utf-16
		    "UTF-16 Unicode coding system (big-endian with BOM)"
		    '(mnemonic "U16-BE-B" bom t little-endian nil))


;; set up unicode mappings, and define fallback charsets

;; First, load the ucs mappings for the existing charsets
;; we need some charsets defined before loading some of these
;; latin 7 is defined elsewhere in latin-euro-latin7, so
;; do that definition now.
;; This of course requires the sumo packages to be already there.
;; Need to fix this.
(require 'latin-euro-latin7)
;; Ditto for latin 8 and 9
(require 'latin-euro-latin8)
(require 'latin-euro-latin9)
(require 'latin-euro-latin10)

;; provide this for lisp use
(defun char-to-ucschar (char &optional no-error)
  "Convert CHAR to a Unicode character. If CHAR is a Mule character,
it is converted according to the `mule-to-ucs' chartab, as set up
by `set-unicode-precedence-list'.
Signals an error if there is no Unicode translation of CHAR, 
unless `no-error' is non-nil, in which case the replacement
character U+FFFD is returned."
  (if (char-mulep char)
      (let ((c (char-ucs char)))
	(if (> c 0)
	    (make-char 'ucs c)
	  (if (null no-error)
	      (signal 'invalid-argument "No Unicode translation of %s" char)
	    (make-char 'ucs #xFFFD))))
    char))


;; convenience function
(defun set-char-ucs-char (char code)
  (set-char-ucs char code)
  (set-ucs-char code char))

(mapcar 'load-library
	'(
	  "u-cns-1.el"
	  "u-cns-2.el"
	  "u-cns-3.el"
	  "u-cns-4.el"
	  "u-cns-5.el"
	  "u-cns-6.el"
	  "u-cns-7.el"
	  "uascii.el" ; also does control-1 and latin-1
	  "ubig5.el"
	  ;; "uethiopic.el" ; not supported, cos we need the charset space
	  "ugb2312.el"
	  ;; "uipa.el" ; mule ipa should not be supported
	  ;; "uiscii.el" ; not in the usual build
	  "uiso8859-13.el"
	  "uiso8859-14.el"
	  "uiso8859-15.el"
	  "uiso8859-16.el"
	  "uiso8859-2.el"
	  "uiso8859-3.el"
	  "uiso8859-4.el"
	  "uiso8859-5.el"
	  "uiso8859-6.el"
	  "uiso8859-7.el"
	  "uiso8859-8.el"
	  "uiso8859-9.el"
	  "ujisx0201.el"
	  "ujisx0208.el"
	  "ujisx0212.el"
	  "uksc5601.el"
	  "usisheng.el"
	  "usupple.el"
	  ;; "utibetan.el" ; not supported in normal build
	  "utis620.el"
	  "uviscii.el"
))

(setq unicode-precedence-list-variable 
      '(ascii latin-iso8859-1
	      latin-iso8859-2 latin-iso8859-3 latin-iso8859-4 cyrillic-iso8859-5
	      greek-iso8859-7 hebrew-iso8859-8 latin-iso8859-9 latin-iso8859-13
	      latin-iso8859-14 latin-iso8859-15 latin-iso8859-16 
	      chinese-big5-1 chinese-big5-2 
	      chinese-gb2312 chinese-cns11643-1 chinese-cns11643-2
	      chinese-cns11643-3 chinese-cns11643-4 chinese-cns11643-5
	      chinese-cns11643-6 chinese-cns11643-7
	      japanese-jisx0208 japanese-jisx0212
	      korean-ksc5601 latin-jisx0201 katakana-jisx0201 thai-tis620 ;; ethiopic
	      vietnamese-viscii-lower vietnamese-viscii-upper))


(defun unicode-precedence-list ()
  "Return the precedence list of charsets for translation from Unicode."
  unicode-precedence-list-variable)

(defun unicode-uniquify-list (head tail)
  (cond ((null tail) (nreverse head))
	(t (unicode-uniquify-list 
	    (if (memq (car tail) head)
		head
	      (cons (car tail) head))
	    (cdr tail)))))


(defun set-unicode-precedence-list (l)
  "Set the precedence list of charsets for translation from Unicode
to the ARG list of charsets (in decreasing priority order)."
  (setq unicode-precedence-list-variable
	(unicode-uniquify-list nil
			       (mapcar (lambda (cs)
				  (cond ((symbolp cs)
					 (if (get-charset cs) cs))
					((charsetp cs)
					 (charset-name cs))
					(error 'wrong-type-argument 'charsetp cs)))
				l)))
  (mapcar
   (lambda (cs)
     (let* ((dim (charset-dimension cs))
	    (chars (charset-chars cs))
	    (start (if (= chars 96) 32 33))
	    (end (if (= chars 96) 127 126))
	    (i start)
	    (j start)
	    u c)
       (if (or (= chars 96) (= chars 94)) ; else not a Mule charset
	   (if (= dim 2)
	       (while (<= i end)
		 (setq j start)
		 (while (<= j end)
		   (setq c (make-char cs i j))
		   (setq u (char-ucs c))
		   (if u (set-ucs-char u c))
		   (setq j (1+ j)))
		 (setq i (1+ i)))
	     (while (<= i end)
	       (setq c (make-char cs i nil))
	       (setq u (char-ucs c))
	       (if u (set-ucs-char u c))
	       (setq i (1+ i)))))))
   (reverse unicode-precedence-list-variable)))

;; here goes...
(set-unicode-precedence-list unicode-precedence-list-variable)

;; no point in leaving this around
(fmakunbound 'unicode-make-fallback-charset)

;; set up the combining characters
(load-library "ucombining.el")
;; set the 2 column property on wide characters
(load-library "ueaw.el")
;; load the bidi classes for unicode
(load-library "ubidi.el")
;; set all legacy r2l charsets to have bidi class R
(mapcar (lambda (c)
	  (if (eq (charset-property c 'direction) 'r2l)
	      (set-char-display-property c 'bidi-class 'bidi-r)))
	(charset-list))

;; compact the property table
(compact-char-display-props)
